/* cpuinfo_tt.c */


#include "cpuinfo_tt.h"

#ifdef __cplusplus
extern "C" {
#endif


typedef struct
{
  DWORD dwL1DataCacheSizeK;
  DWORD dwL1InstCacheSizeK;
  DWORD dwL2CacheSizeK;
  DWORD dwL3CacheSizeK;
} CPU_CACHE_SIZE_K;


void GetCacheSizeK_from_Descriptor_Component(int CPUInfo_reg, BOOL bEAX, CPU_SIGNATURE * psig, CPU_CACHE_SIZE_K * pccs);
void GetCacheSizeK_from_Descriptor(DWORD Descriptor, CPU_SIGNATURE * psig, CPU_CACHE_SIZE_K * pccs);
BOOL StreamingSIMDExtensionOSSupport();


void GetCacheSizeK_from_Descriptor_Component(int CPUInfo_reg,
                                             BOOL bEAX,
                                             CPU_SIGNATURE * psig,
                                             CPU_CACHE_SIZE_K * pccs)
{
  if ((CPUInfo_reg & (1 << 31)) == 0)
  {
    GetCacheSizeK_from_Descriptor((CPUInfo_reg >> 24) & 0xFF, psig, pccs);
    GetCacheSizeK_from_Descriptor((CPUInfo_reg >> 16) & 0xFF, psig, pccs);
    GetCacheSizeK_from_Descriptor((CPUInfo_reg >>  8) & 0xFF, psig, pccs);
    if (!bEAX)
    {
      GetCacheSizeK_from_Descriptor((CPUInfo_reg & 0xFF), psig, pccs);
    }
  }
}


void GetCacheSizeK_from_Descriptor(DWORD Descriptor,
                                   CPU_SIGNATURE * psig,
                                   CPU_CACHE_SIZE_K * pccs)
{
  if (lstrcmp(psig->szVendorID, "GenuineIntel") == 0)
  {
    /* Intel */
    switch (Descriptor)
    {
    case 0x06:
      /* 1st-level instruction cache: 8-KB, 4-way set associative, 32-byte line size */
      pccs->dwL1InstCacheSizeK = 8;
      break;

    case 0x08:
      /* 1st-level instruction cache: 16-KB, 4-way set associative, 32-byte line size */
      pccs->dwL1InstCacheSizeK = 16;
      break;

    case 0x0A:
      /* 1st-level data cache: 8-KB, 2-way set associative, 32-byte line size */
      pccs->dwL1DataCacheSizeK = 8;
      break;

    case 0x0C:
      /* 1st-level data cache: 16-KB, 4-way set associative, 32-byte line size */
      pccs->dwL1DataCacheSizeK = 16;
      break;

    case 0x22:
      /* 3rd-level cache: 512 KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL3CacheSizeK = 512;
      break;

    case 0x23:
      /* 3rd-level cache: 1-MB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 1;
      break;

    case 0x25:
      /* 3rd-level cache: 2-MB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 2;
      break;

    case 0x29:
      /* 3rd-level cache: 4-MB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 4;
      break;

    case 0x2C:
      /* 1st-level data cache: 32-KB, 8-way set associative, 64-byte line size */
      pccs->dwL1DataCacheSizeK = 32;
      break;

    case 0x30:
      /* 1st-level instruction cache: 32-KB, 8-way set associative, 64-byte line size */
      pccs->dwL1InstCacheSizeK = 32;
      break;

    case 0x39:
      /* 2nd-level cache: 128-KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 128;
      break;

    case 0x3A:
      /* 2nd-level cache: 192-KB, 6-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 192;
      break;

    case 0x3B:
      /* 2nd-level cache: 128-KB, 2-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 128;
      break;

    case 0x3C:
      /* 2nd-level cache: 256-KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 256;
      break;

    case 0x3D:
      /* 2nd-level cache: 384-KB, 6-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 384;
      break;

    case 0x3E:
      /* 2nd-level cache: 512-KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x41:
      /* 2nd-level cache: 128-KB, 4-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 128;
      break;

    case 0x42:
      /* 2nd-level cache: 256-KB, 4-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 256;
      break;

    case 0x43:
      /* 2nd-level cache: 512-KB, 4-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x44:
      /* 2nd-level cache: 1-MB, 4-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 1;
      break;

    case 0x45:
      /* 2nd-level cache: 2-MB, 4-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 2;
      break;

    case 0x46:
      /* 3rd-level cache: 4MB, 4-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 4;
      break;

    case 0x47:
      /* 3rd-level cache: 8MB, 8-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 8;
      break;

    case 0x49:
      if (0x0F == psig->dwFamily && 0x06 == psig->dwModel)
      {
        /* 3rd-level cache: 4MB, 16-way set associative, 64-byte line size (Intel Xeon processor MP) */
        pccs->dwL3CacheSizeK = 1024 * 4;
      }
      else
      {
        /* 2nd-level cache: 4MB, 16-way set associative, 64-byte line size */
        pccs->dwL2CacheSizeK = 1024 * 4;
      }
      break;

    case 0x4A:
      /* 3rd-level cache: 6MB, 12-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 6;
      break;

    case 0x4B:
      /* 3rd-level cache: 8MB, 16-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 8;
      break;

    case 0x4C:
      /* 3rd-level cache: 12MB, 12-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 12;
      break;

    case 0x4D:
      /* 3rd-level cache: 16MB, 16-way set associative, 64-byte line size */
      pccs->dwL3CacheSizeK = 1024 * 16;
      break;

    case 0x4E:
      /* 2nd-level cache: 6MB, 24-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 6;
      break;

    case 0x60:
      /* 1st-level data cache: 16-KB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL1DataCacheSizeK = 16;
      break;

    case 0x66:
      /* 1st-level data cache: 8-KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL1DataCacheSizeK = 8;
      break;

    case 0x67:
      /* 1st-level data cache: 16-KB, 4-way set associative, sectored cache, 64-byte line size */
      pccs->dwL1DataCacheSizeK = 16;
      break;

    case 0x68:
      /* 1st-level data cache: 32-KB, 4 way set associative, sectored cache, 64-byte line size */
      pccs->dwL1DataCacheSizeK = 32;
      break;

    case 0x78:
      /* 2nd-level cache: 1-MB, 4-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 1;
      break;

    case 0x79:
      /* 2nd-level cache: 128-KB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 128;
      break;

    case 0x7A:
      /* 2nd-level cache: 256-KB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 256;
      break;

    case 0x7B:
      /* 2nd-level cache: 512-KB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x7C:
      /* 2nd-level cache: 1-MB, 8-way set associative, sectored cache, 64-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 1;
      break;

    case 0x7D:
      /* 2nd-level cache: 2-MB, 8-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 2;
      break;

    case 0x7F:
      /* 2nd-level cache: 512-KB, 2-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x82:
      /* 2nd-level cache: 256-KB, 8-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 256;
      break;

    case 0x83:
      /* 2nd-level cache: 512-KB, 8-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x84:
      /* 2nd-level cache: 1-MB, 8-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 1;
      break;

    case 0x85:
      /* 2nd-level cache: 2-MB, 8-way set associative, 32-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 2;
      break;

    case 0x86:
      /* 2nd-level cache: 512-KB, 4-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 512;
      break;

    case 0x87:
      /* 2nd-level cache: 1-MB, 8-way set associative, 64-byte line size */
      pccs->dwL2CacheSizeK = 1024 * 1;
      break;
    }
  }
}


BOOL StreamingSIMDExtensionOSSupport()
{
  __try
  {
    __asm
    {
      xorps xmm1, xmm2  /* this is an instruction available in SSE */
    }
  }
  __except (EXCEPTION_EXECUTE_HANDLER)
  {
    if (_exception_code() == STATUS_ILLEGAL_INSTRUCTION)
    {
      return FALSE;
    }
    return FALSE;
  }
  return TRUE;
}


void GetCpuVendorAndSignature(CPU_SIGNATURE * psig)
{
  static CPU_SIGNATURE sig = {0};

  if (0 == sig.szVendorID[0])
  {
    CPU_INFO CPUInfo;
    DWORD dwLargestStdFunc = 0;   /* Largest Standard Function */
    DWORD dwLargestExtFunc = 0;   /* Largest Extended Function */

    /* Vendor-ID and Largest Standard Function (Function 0) */
    __cpuid((int*)&CPUInfo, 0);
    dwLargestStdFunc = CPUInfo.EAX;

    /* Vendor-ID */
    *((int*)&sig.szVendorID[0]) = CPUInfo.EBX;
    *((int*)&sig.szVendorID[4]) = CPUInfo.EDX;
    *((int*)&sig.szVendorID[8]) = CPUInfo.ECX;

    if (dwLargestStdFunc >= 1)
    {
      /* Feature Information (Function 1) */
      __cpuid((int*)&CPUInfo, 1);

      /* Processor Signature */
      sig.dwFamily =
        ((CPUInfo.EAX >> 20) & 0xFF) +
        ((CPUInfo.EAX >> 8) & 0xF);

      sig.dwModel =
        (((CPUInfo.EAX >> 16) & 0xF) << 4) + 
        ((CPUInfo.EAX >> 4) & 0xF);

      sig.dwStepping = (CPUInfo.EAX) & 0xF;

      /* brand ID */
      sig.dwBrandID = (CPUInfo.EBX & 0xFF);
    }

    /* Largest Extended Function */
    __cpuid((int*)&CPUInfo, 0x80000000);
    dwLargestExtFunc = CPUInfo.EAX;

    if (dwLargestExtFunc >= 0x80000001)
    {
      /* extended brand ID */
      sig.dwExtBrandID = (CPUInfo.EBX & 0xFFFF);
    }

    if (dwLargestExtFunc >= 0x80000004)
    {
      /* Processor Name String Identifier */
      int i;

      for (i = 0; i <= 2; i++)
      {
        __cpuid((int*)&CPUInfo, 0x80000002 + i);
        __movsd((LPDWORD)&sig.szNameID[i * 16], (LPDWORD)&CPUInfo, 4);
      }
    }
  }

  *psig = sig;
}


#define CPU_FEATURE_BT(reg, bit) ((CPUInfo.reg) & (1 << (bit)))

/* get CPU feature */
DWORD GetCpuFeature()
{
  static DWORD dwCpuFeature = 0;

  if (0 == dwCpuFeature)
  {
    CPU_INFO CPUInfo;
    DWORD dwLargestStdFunc = 0;   /* Largest Standard Function */
    DWORD dwLargestExtFunc = 0;   /* Largest Extended Function */
    BOOL bOsSupportXMM = StreamingSIMDExtensionOSSupport();

    /* largest function */
    __cpuid((int*)&CPUInfo, 0);
    dwLargestStdFunc = CPUInfo.EAX;
    __cpuid((int*)&CPUInfo, 0x80000000);
    dwLargestExtFunc = CPUInfo.EAX;

    if (dwLargestStdFunc >= 1)
    {
      __cpuid((int*)&CPUInfo, 1);

      if (CPU_FEATURE_BT(EDX, 23))
      {
        /* MMX */
        dwCpuFeature |= _CPU_FEATURE_MMX;
      }
      if (CPU_FEATURE_BT(EDX, 15))
      {
        /* CMOV */
        dwCpuFeature |= _CPU_FEATURE_CMOV;
      }

      if (bOsSupportXMM)
      {
        if (CPU_FEATURE_BT(EDX, 25))
        {
          /* SSE */
          dwCpuFeature |= _CPU_FEATURE_SSE;
        }
        if (CPU_FEATURE_BT(EDX, 26))
        {
          /* SSE2*/
          dwCpuFeature |= _CPU_FEATURE_SSE2;
        }
        if (CPU_FEATURE_BT(ECX, 0))
        {
          /* SSE3 */
          dwCpuFeature |= _CPU_FEATURE_SSE3;
        }
        if (CPU_FEATURE_BT(ECX, 9))
        {
          /* SSSE3 */
          dwCpuFeature |= _CPU_FEATURE_SSSE3;
        }
        if (CPU_FEATURE_BT(ECX, 19))
        {
          /* SSE4.1 */
          dwCpuFeature |= _CPU_FEATURE_SSE4_1;
        }
        if (CPU_FEATURE_BT(ECX, 20))
        {
          /* SSE4.2 */
          dwCpuFeature |= _CPU_FEATURE_SSE4_2;
        }
      }
    }

    if (dwLargestExtFunc >= 0x80000001)
    {
      __cpuid((int*)&CPUInfo, 0x80000001);

      if (CPU_FEATURE_BT(EDX, 22))
      {
        /* MMX extention */
        dwCpuFeature |= _CPU_FEATURE_MMXE;
      }
      if (CPU_FEATURE_BT(EDX, 31))
      {
        /* 3DNow */
        dwCpuFeature |= _CPU_FEATURE_3DNow;
      }
      if (CPU_FEATURE_BT(EDX, 30))
      {
        /* Enhanced 3DNow */
        dwCpuFeature |= _CPU_FEATURE_3DNowE;
      }

      if (bOsSupportXMM)
      {
        if (CPU_FEATURE_BT(ECX, 6))
        {
          /* SSE4A */
          dwCpuFeature |= _CPU_FEATURE_SSE4A;
        }
        if (CPU_FEATURE_BT(ECX, 11))
        {
          /* SSE5 */
          dwCpuFeature |= _CPU_FEATURE_SSE5;
        }
      }
    }
  }
  return dwCpuFeature;
}


/* get minimum non-temporal data size */
DWORD GetNonTemporalDataSizeMin()
{
  static DWORD dwNonTemporalDataSizeMin = 0;

  if (0 == dwNonTemporalDataSizeMin)
  {
    DWORD dwCpuFeature = GetCpuFeature();

    dwNonTemporalDataSizeMin = NON_TEMPORAL_STORES_NOT_SUPPORTED;

    if ((dwCpuFeature & _CPU_FEATURE_MMXE) ||
        (dwCpuFeature & _CPU_FEATURE_SSE))
    {
      /*
       Intel (Pentium 3 or higher, SSE)
       AMD (K7 or higher, MMX+ or SSE)
       VIA (C3 Nehemiah or higher, SSE)
       Transmeta (Efficeon or higher, SSE)
      */

      CPU_INFO CPUInfo;
      DWORD dwLargestStdFunc = 0;   /* Largest Standard Function */
      DWORD dwLargestExtFunc = 0;   /* Largest Extended Function */
      CPU_CACHE_SIZE_K ccs = {0};   /* cache size */
      CPU_SIGNATURE sig;            /* VendorID and Signature */

      __stosb((unsigned char *)&sig, 0, sizeof(sig));

      /* VendorID and Signature */
      GetCpuVendorAndSignature(&sig);

      /* largest function */
      __cpuid((int*)&CPUInfo, 0);
      dwLargestStdFunc = CPUInfo.EAX;
      __cpuid((int*)&CPUInfo, 0x80000000);
      dwLargestExtFunc = CPUInfo.EAX;

      if (dwLargestStdFunc >= 2)
      {
        /* Cache Descriptors (Function 2) */
        DWORD dwReqCpuidCount = 0;

        __cpuid((int*)&CPUInfo, 2);
        dwReqCpuidCount = (CPUInfo.EAX & 0xFF);
        while (dwReqCpuidCount-- >= 1)
        {
          GetCacheSizeK_from_Descriptor_Component(CPUInfo.EAX, TRUE,  &sig, &ccs);
          GetCacheSizeK_from_Descriptor_Component(CPUInfo.EBX, FALSE, &sig, &ccs);
          GetCacheSizeK_from_Descriptor_Component(CPUInfo.ECX, FALSE, &sig, &ccs);
          GetCacheSizeK_from_Descriptor_Component(CPUInfo.EDX, FALSE, &sig, &ccs);
          __cpuid((int*)&CPUInfo, 2);
        }
      }


      if (dwLargestExtFunc >= 0x80000006)
      {
        DWORD dwL1DataCacheSizeK = 0;
        DWORD dwL1InstCacheSizeK = 0;
        DWORD dwL2CacheSizeK = 0;
        DWORD dwL3CacheSizeK = 0;

        /* L1 Cache */
        __cpuid((int*)&CPUInfo, 0x80000005);
        dwL1DataCacheSizeK = CPUInfo.ECX >> 24;
        if (dwL1DataCacheSizeK > 0)
        {
          ccs.dwL1DataCacheSizeK = dwL1DataCacheSizeK;
        }
        dwL1InstCacheSizeK = CPUInfo.EDX >> 24;
        if (dwL1InstCacheSizeK > 0)
        {
          ccs.dwL1InstCacheSizeK = dwL1InstCacheSizeK;
        }

        /* L2/L3 Cache */
        __cpuid((int*)&CPUInfo, 0x80000006);
        dwL2CacheSizeK = CPUInfo.ECX >> 16;
        if (dwL2CacheSizeK > 0)
        {
          if (1 == dwL2CacheSizeK)
          {
            /* AMD K7 with CPUID=0630h (Duron) */
            /* inadvertently report 1 KB instead of 64 KB. */
            dwL2CacheSizeK = 64;
          }
          ccs.dwL2CacheSizeK = dwL2CacheSizeK;
        }
        dwL3CacheSizeK = ((CPUInfo.EDX & 0xFFFC0000) >> 18) * 512;
        if (dwL3CacheSizeK > 0)
        {
          ccs.dwL3CacheSizeK = dwL3CacheSizeK;
        }
      }

      dwNonTemporalDataSizeMin = max(ccs.dwL1DataCacheSizeK * 1024,
        max(ccs.dwL2CacheSizeK * 1024, ccs.dwL3CacheSizeK * 1024));
    }
  }

  return dwNonTemporalDataSizeMin;
}


#ifdef __cplusplus
}
#endif
