
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <intrin.h>
#include "cpuinfo_tt.h"

__declspec(noinline)
__int64 rdtsc() {
  __asm {
    cpuid
    rdtsc
  }
};

#define REP_COUNT 2000

__declspec(noinline)
__int64 floor_test()
{
  __int64 start, stop;
  double d = 0.0;

  start = rdtsc();
  for (int i = 0; i < REP_COUNT; i++)
  {
    d += floor(i + 0.1);
  }
  stop = rdtsc();

  printf("", d);
  return (stop - start) / REP_COUNT;
}

__declspec(noinline)
__int64 ceil_test()
{
  __int64 start, stop;
  double d = 0.0;

  start = rdtsc();
  for (int i = 0; i < REP_COUNT; i++)
  {
    d += ceil(i + 0.1);
  }
  stop = rdtsc();

  printf("", d);
  return (stop - start) / REP_COUNT;
}

__int64 DoFloorBench()
{
  __int64 result[3] = {0};

  do {
    result[0] = floor_test();
    result[1] = floor_test();
    result[2] = floor_test();
  }
  while (result[0] != result[1] || result[1] != result[2]);

  return result[0];
}

__int64 DoCeilBench()
{
  __int64 result[3] = {0};

  do {
    result[0] = ceil_test();
    result[1] = ceil_test();
    result[2] = ceil_test();
  }
  while (result[0] != result[1] || result[1] != result[2]);

  return result[0];
}



int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow)
{
  char sz[2048];
  char szTemp[1024];


  sz[0] = 0;
  szTemp[0] = 0;

  CPU_SIGNATURE sig;
  GetCpuVendorAndSignature(&sig);
  if (sig.szNameID[0] != 0)
  {
    sprintf(szTemp, "%s\n\n", sig.szNameID);
    strcat(sz, szTemp);
  }
  else
  {
    if (lstrcmp(sig.szVendorID, "GenuineIntel") == 0)
    {
      static const char * pszIntelBrands[256] =
      {
        0,
        "Intel Celeron processor",
        "Intel Pentium III processor",
        "Intel Pentium III Xeon processor",   // If processor signature = 000006B1h, then gIntel Celeron processorh
        "Intel Pentium III processor",
        0,
        "Mobile Intel Pentium III Processor-M",
        "Mobile Intel Celeron processor",
        "Intel Pentium 4 processor",          // If processor signature is >=00000F13h, then gIntel Genuine processorh
        "Intel Pentium 4 processor",
        "Intel Celeron Processor",
        "Intel Xeon processor",               // If processor signature is <00000F13h, then gIntel Xeon processor MPh
        "Intel Xeon processor MP",
        0,
        "Mobile Intel Pentium 4 processor-M", // If processor signature is <00000F13h, then gIntel Xeon processorh
        "Mobile Intel Celeron processor",
        "Mobile Genuine Intel processor",
        "Intel Celeron M processor",
        "Mobile Intel Celeron processor",
        "Intel Celeron Processor",
        "Mobile Genuine Intel processor",
        "Intel Pentium M processor",
        "Mobile Intel Celeron processor",
      };

      if (pszIntelBrands[sig.dwBrandID] != 0)
      {
        const char * pszIntelBrand = pszIntelBrands[sig.dwBrandID];
        CPU_INFO CPUInfo;

        __cpuid((int*)&CPUInfo, 1);
        switch (sig.dwBrandID)
        {
        case 0x03:
          if (CPUInfo.EAX == 0x00006B1)
          {
            pszIntelBrand = "Intel Celeron processor";
          }
          break;

        case 0x08:
          if (CPUInfo.EAX >= 0x00000F13)
          {
            pszIntelBrand = "Intel Genuine processor";
          }
          break;

        case 0x0B:
          if (CPUInfo.EAX < 0x00000F13)
          {
            pszIntelBrand = "Intel Xeon processor MP";
          }
          break;

        case 0x0E:
          if (CPUInfo.EAX < 0x00000F13)
          {
            pszIntelBrand = "Intel Xeon processor";
          }
          break;
        }
        sprintf(szTemp, "%s\n\n", pszIntelBrand);
        strcat(sz, szTemp);
      }
    }
  }



  SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
  SetThreadAffinityMask(GetCurrentThread(), 1);



  _set_SSE2_enable(0);
  sprintf(szTemp, "floor (disable sse2):\t%I64d\n", DoFloorBench());
  strcat(sz, szTemp);

  _set_SSE2_enable(1);
  sprintf(szTemp, "floor (normal):\t%I64d\n", DoFloorBench());
  strcat(sz, szTemp);

  strcat(sz, "\n");

  _set_SSE2_enable(0);
  sprintf(szTemp, "ceil (disable sse2):\t%I64d\n", DoCeilBench());
  strcat(sz, szTemp);

  _set_SSE2_enable(1);
  sprintf(szTemp, "ceil (normal):\t%I64d\n", DoCeilBench());
  strcat(sz, szTemp);



  MessageBox(NULL, sz, "floor & ceil benchmark (mozcrt19.dll)", 0);

  return 0;
}